@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2021 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -17,6 +17,7 @@
 /* setscreen operator for Ghostscript library */
 #include "memory_.h"
 #include "string_.h"
+#include "assert_.h"
 #include <stdlib.h>             /* for qsort */
 #include "gx.h"
 #include "gserrors.h"
@@ -27,6 +28,9 @@
 #include "gxdevice.h"           /* for gzht.h */
 #include "gzht.h"
 #include "gxfmap.h"             /* For effective transfer usage in threshold */
+#include "gp.h"
+
+#define DUMP_SCREENS 0
 
 /* Forward declarations */
 void gx_set_effective_transfer(gs_gstate *);
@@ -165,12 +169,12 @@ gs_currentscreenlevels(const gs_gstate * pgs)
 {
     int gi = 0;
 
-    if (pgs->device != 0)
+    if (pgs->device != NULL)
         gi = pgs->device->color_info.gray_index;
     if (gi != GX_CINFO_COMP_NO_INDEX)
-        return pgs->dev_ht->components[gi].corder.num_levels;
+        return pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[gi].corder.num_levels;
     else
-        return pgs->dev_ht->components[0].corder.num_levels;
+        return pgs->dev_ht[HT_OBJTYPE_DEFAULT]->components[0].corder.num_levels;
 }
 
 /* .setscreenphase */
@@ -318,11 +322,11 @@ gx_ht_copy_ht_order(gx_ht_order * pdest, gx_ht_order * psrc, gs_memory_t * mem)
                      psrc->procs, mem);
     if (code < 0)
         return code;
-    if (pdest->levels != 0)
+    if (pdest->levels != NULL)
         memcpy(pdest->levels, psrc->levels, psrc->num_levels * sizeof(uint));
-    if (pdest->bit_data != 0)
+    if (pdest->bit_data != NULL)
         memcpy(pdest->bit_data, psrc->bit_data,
-                psrc->num_bits * psrc->procs->bit_data_elt_size);
+               (size_t)psrc->num_bits * psrc->procs->bit_data_elt_size);
     pdest->transfer = psrc->transfer;
     rc_increment(pdest->transfer);
     return 0;
@@ -384,12 +388,20 @@ gx_ht_alloc_threshold_order(gx_ht_order * porder, uint width, uint height,
                             uint num_levels, gs_memory_t * mem)
 {
     gx_ht_order order;
-    uint num_bits = width * height;
-    const gx_ht_order_procs_t *procs =
-        (num_bits > 2000 && num_bits <= max_ushort ?
-         &ht_order_procs_short : &ht_order_procs_default);
+
+    unsigned long num_bits = bitmap_raster(width) * (unsigned long)8 * height;
+    const gx_ht_order_procs_t *procs;
     int code;
 
+    if (num_bits <= 2000)
+        procs = &ht_order_procs_default;
+    else if (num_bits <= max_ushort + 1)  /* We can index 0 to 65535 so a size of 65536 (max_ushort + 1) is OK */
+        procs = &ht_order_procs_short;
+    else if (num_bits <= max_uint)
+        procs = &ht_order_procs_uint;
+    else
+        return_error(gs_error_VMerror);  /* At this point in history, this is way too large of a screen */
+
     order = *porder;
     gx_compute_cell_values(&order.params);
     code = gx_ht_alloc_ht_order(&order, width, height, num_levels,
@@ -447,7 +459,6 @@ gx_sort_ht_order(gx_ht_bit * recs, uint N)
     for (i = 0; i < N; i++)
         recs[i].offset = i;
     qsort((void *)recs, N, sizeof(*recs), compare_samples);
-#ifndef GS_THREADSAFE
 #ifdef DEBUG
     if (gs_debug_c('H')) {
         uint i;
@@ -458,7 +469,6 @@ gx_sort_ht_order(gx_ht_bit * recs, uint N)
                       i, recs[i].offset, recs[i].mask);
     }
 #endif
-#endif
 }
 
 /*
@@ -548,7 +558,7 @@ gx_ht_construct_bits(gx_ht_order * porder)
         gx_ht_construct_bit(phb, porder->width, phb->offset);
 #ifdef DEBUG
     if (gs_debug_c('H')) {
-        dmlprintf1(porder->data_memory, "[H]Halftone order bits 0x%lx:\n", (ulong)porder->bit_data);
+        dmlprintf1(porder->data_memory, "[H]Halftone order bits "PRI_INTPTR":\n", (intptr_t)porder->bit_data);
         for (i = 0, phb = (gx_ht_bit *)porder->bit_data;
              i < porder->num_bits;
              i++, phb++)
@@ -565,13 +575,13 @@ gx_ht_order_release(gx_ht_order * porder, gs_memory_t * mem, bool free_cache)
 {
     /* "free cache" is a proxy for "differs from default" */
     if (free_cache) {
-        if (porder->cache != 0)
+        if (porder->cache != NULL)
             gx_ht_free_cache(mem, porder->cache);
     }
     porder->cache = 0;
     rc_decrement(porder->transfer, "gx_ht_order_release(transfer)");
     porder->transfer = 0;
-    if (porder->data_memory != 0) {
+    if (porder->data_memory != NULL) {
         gs_free_object(porder->data_memory, porder->bit_data,
                        "gx_ht_order_release(bit_data)");
         gs_free_object(porder->data_memory, porder->levels,
@@ -636,10 +646,10 @@ gs_color_name_component_number(gx_device * dev, const char * pname,
     int num_colorant;
 
 #define check_colorant_name(dev, name) \
-    ((*dev_proc(dev, get_color_comp_index)) (dev, name, strlen(name), NO_COMP_NAME_TYPE))
+    ((*dev_proc(dev, get_color_comp_index)) (dev, name, strlen(name), NO_COMP_NAME_TYPE_HT))
 
 #define check_colorant_name_length(dev, name, length) \
-    ((*dev_proc(dev, get_color_comp_index)) (dev, name, length, NO_COMP_NAME_TYPE))
+    ((*dev_proc(dev, get_color_comp_index)) (dev, name, length, NO_COMP_NAME_TYPE_HT))
 
 #define check_name(str, pname, length) \
     ((strlen(str) == length) && (strncmp(pname, str, length) == 0))
@@ -938,7 +948,8 @@ gx_gstate_dev_ht_install(
     gs_gstate *       pgs,
     gx_device_halftone *    pdht,
     gs_halftone_type        type,
-    const gx_device *       dev )
+    const gx_device *       dev,
+    gs_HT_objtype_t objtype )
 {
     gx_device_halftone      dht;
     int                     num_comps = pdht->num_dev_comp;
@@ -949,6 +960,8 @@ gx_gstate_dev_ht_install(
     uint w, h;
     int dw, dh;
 
+    assert(objtype < HT_OBJTYPE_COUNT);
+
     /* construct the new device halftone structure */
     memset(&dht.order, 0, sizeof(dht.order));
     /* the rc field is filled in later */
@@ -977,7 +990,7 @@ gx_gstate_dev_ht_install(
      * by clearing the corresponding pointers in the operand halftone's
      * orders.
      */
-    if (pdht->components != 0) {
+    if (pdht->components != NULL) {
         int     input_ncomps = pdht->num_comp;
 
         for (i = 0; i < input_ncomps && code >= 0; i++) {
@@ -985,7 +998,8 @@ gx_gstate_dev_ht_install(
             gx_ht_order *           p_s_order = &p_s_comp->corder;
             int                     comp_num = p_s_comp->comp_number;
 
-            if (comp_num >= 0 && comp_num < GX_DEVICE_COLOR_MAX_COMPONENTS) {
+            if (comp_num >= 0 && comp_num < GX_DEVICE_COLOR_MAX_COMPONENTS &&
+                comp_num < dht.num_comp) {
                 gx_ht_order *   p_d_order = &dht.components[comp_num].corder;
 
                 /* indicate that this order has been filled in */
@@ -1089,27 +1103,28 @@ gx_gstate_dev_ht_install(
      * we still need.
      */
     if (code >= 0) {
-        gx_device_halftone *    pgsdht = pgs->dev_ht;
-        rc_header               tmp_rc;
+        gx_device_halftone **ppgsdht;
+        rc_header tmp_rc;
 
-        if (pgsdht != 0 && pgsdht->rc.ref_count == 1) {
-            if (pdht != pgsdht)
-                gx_device_halftone_release(pgsdht, pgsdht->rc.memory);
+        /* The pgsdht corresponds to the one we will be installing according to 'objtype' */
+        ppgsdht = &(pgs->dev_ht[objtype]);
+        if (*ppgsdht != NULL && (*ppgsdht)->rc.ref_count == 1) {
+             if (pdht != *ppgsdht)
+                gx_device_halftone_release(*ppgsdht, (*ppgsdht)->rc.memory);
         } else {
-            rc_unshare_struct( pgs->dev_ht,
+            rc_unshare_struct( *ppgsdht,
                                gx_device_halftone,
                                &st_device_halftone,
                                pgs->memory,
                                BEGIN code = gs_error_VMerror; goto err; END,
                                "gx_gstate_dev_ht_install" );
-            pgsdht = pgs->dev_ht;
         }
 
         /*
          * Everything worked. "Assume ownership" of the appropriate
          * portions of the source device halftone by clearing the
          * associated references.  Since we might have
-         * pdht == pgs->dev_ht, this must done before updating pgs->dev_ht.
+         * pdht == pgs->dev_ht[], this must done before updating pgs->dev_ht[].
          *
          * If the default order has been used for a device component, and
          * any of the source component orders share their levels or bit_data
@@ -1118,7 +1133,7 @@ gx_gstate_dev_ht_install(
          * be cleared immediately below, so subsequently it will not be
          * possible to tell if that this information is being shared.
          */
-        if (pdht->components != 0) {
+        if (pdht->components != NULL) {
             int     input_ncomps = pdht->num_comp;
 
             for (i = 0; i < input_ncomps; i++) {
@@ -1138,9 +1153,9 @@ gx_gstate_dev_ht_install(
             memset(&pdht->order, 0, sizeof(pdht->order));
         }
 
-        tmp_rc = pgsdht->rc;
-        *pgsdht = dht;
-        pgsdht->rc = tmp_rc;
+        tmp_rc = (*ppgsdht)->rc;
+        **ppgsdht = dht;
+        (*ppgsdht)->rc = tmp_rc;
 
         /* update the effective transfer function array */
         gx_gstate_set_effective_xfer(pgs);
@@ -1167,6 +1182,22 @@ gx_gstate_dev_ht_install(
     return code;
 }
 
+/*
+ * Copy the dev_ht[HT_OBJTYPE_DEFAULT] to the dev_ht[] for the specified object type.
+ */
+int
+gx_gstate_dev_ht_copy_to_objtype(gs_gstate *pgs, gs_HT_objtype_t objtype)
+{
+    gx_device_halftone *pdht = pgs->dev_ht[HT_OBJTYPE_DEFAULT];	/* the current dev_ht */
+
+    if (objtype >= HT_OBJTYPE_COUNT) {
+        return_error(gs_error_undefined);
+    }
+    rc_increment(pdht);
+    pgs->dev_ht[objtype] = pdht;
+    return 0;
+}
+
 /*
  * Install a new halftone in the graphics state.  Note that we copy the top
  * level of the gs_halftone and the gx_device_halftone, and take ownership
@@ -1182,7 +1213,7 @@ gx_ht_install(gs_gstate * pgs, const gs_halftone * pht,
     int code;
 
     pdht->num_dev_comp = pgs->device->color_info.num_components;
-    if (old_ht != 0 && old_ht->rc.memory == mem &&
+    if (old_ht != NULL && old_ht->rc.memory == mem &&
         old_ht->rc.ref_count == 1
         )
         new_ht = old_ht;
@@ -1191,7 +1222,8 @@ gx_ht_install(gs_gstate * pgs, const gs_halftone * pht,
                           mem, return_error(gs_error_VMerror),
                           "gx_ht_install(new halftone)");
     code = gx_gstate_dev_ht_install(pgs,
-                             pdht, pht->type, gs_currentdevice_inline(pgs));
+                             pdht, pht->type, gs_currentdevice_inline(pgs),
+                             pht->objtype);
     if (code < 0) {
         if (new_ht != old_ht)
             gs_free_object(mem, new_ht, "gx_ht_install(new halftone)");
@@ -1230,7 +1262,7 @@ gx_ht_install(gs_gstate * pgs, const gs_halftone * pht,
 void
 gx_gstate_set_effective_xfer(gs_gstate * pgs)
 {
-    gx_device_halftone *pdht = pgs->dev_ht;
+    gx_device_halftone *pdht = pgs->dev_ht[HT_OBJTYPE_DEFAULT];
     gx_transfer_map *pmap;
     gx_ht_order *porder;
     int i, component_num, non_id_count;
@@ -1271,7 +1303,10 @@ gx_gstate_set_effective_xfer(gs_gstate * pgs)
         }
     }
 
-    if (pdht) { /* might not be initialized yet */
+    /* HT may not be initialized yet.  Only do if the target is a halftone device.
+       Per the spec, the HT is a self-contained description of a halftoning process.
+       We don't use any xfer function from the HT if we are not halftoning */
+    if (pdht && !device_is_contone(pgs->device)) {
 
         /* Since the transfer function is pickled into the threshold array (if any)*/
         /*  we need to free it so it can be reconstructed with the current transfer */
@@ -1363,7 +1398,7 @@ gx_ht_construct_threshold( gx_ht_order *d_order, gx_device *dev,
     shift = d_order->shift;
 
     if (d_order->threshold != NULL) return 0;
-    thresh = (byte *)gs_malloc(memory, d_order->width * d_order->full_height, 1,
+    thresh = (byte *)gs_malloc(memory, (size_t)d_order->width * d_order->full_height, 1,
                               "gx_ht_construct_threshold");
     if (thresh == NULL) {
         return -1 ;         /* error if allocation failed   */
@@ -1439,13 +1474,34 @@ gx_ht_construct_threshold( gx_ht_order *d_order, gx_device *dev,
     }
 #ifdef DEBUG
     if ( gs_debug_c('h') ) {
-         for( i=0; i<(int)d_order->height; i++ ) {
-            dmprintf1(memory, "threshold array row %3d= ", i);
-            for( j=0; j<(int)(d_order->width); j++ )
-                dmprintf1(memory, "%3d ", *(thresh+j+(i*d_order->width)) );
-            dmprintf(memory, "\n");
+         dmprintf3(memory, "threshold array component %d [ %d x %d ]:\n",
+                  plane_index, (int)(d_order->full_height), (int)(d_order->width));
+         for( i=0; i<(int)d_order->full_height; i++ ) {
+            dmprintf1(memory, "row %3d= ", i);
+            for( j=0; j<(int)(d_order->width); j++ ) {
+                dmprintf1(memory, "%02x ", *(thresh+j+(i*d_order->width)) );
+                if ((j&31) == 31)
+                    dmprintf(memory, "\n         ");
+            }
+            if ((j&31) != 0)
+                dmprintf(memory, "\n");
         }
    }
 #endif
+/* Large screens are easier to see as images */
+#if DUMP_SCREENS
+    {
+        char file_name[50];
+        gp_file *fid;
+
+        snprintf(file_name, 50, "Screen_From_Tiles_%dx%d.raw", d_order->width, d_order->full_height);
+        fid = gp_fopen(memory, file_name, "wb");
+        if (fid) {
+            gp_fwrite(thresh, sizeof(unsigned char), d_order->width * d_order->full_height, fid);
+            gp_fclose(fid);
+        }
+    }
+#endif
+
     return 0;
 }
